'use strict';

const shiki = require('shiki');

function addClass(code, classes, tag) {
  const classRE = new RegExp(`<${tag ?? "w+"}[^>]*class="([\\w+-:;\\/* ]*)"`);
  classes = Array.isArray(classes) ? classes : [classes];
  return code.replace(classRE, (match, previousClasses) => {
    return match.replace(previousClasses, `${previousClasses} ${classes}`);
  });
}

function createRangeTransformer(classMap, options = {}) {
  return ({ code }) => {
    const tagRE = options.tagRegExp ?? /(?:\/\/|\/\*{1,2}) *\[!code ([\w+-]+)(?::(\d+))?] *(?:\*{1,2}\/)?/;
    const lineOptions = [];
    const tags = Object.keys(classMap);
    code = code.split("\n").map((lineOfCode, lineNumber) => {
      const [match, tag, range] = lineOfCode.match(tagRE) ?? [];
      if (!match) {
        return lineOfCode;
      }
      if (!tags.includes(tag)) {
        return lineOfCode;
      }
      for (const [rangeOffset] of Array.from({ length: Number(range ?? 1) }).entries()) {
        lineOptions.push({
          line: lineNumber + rangeOffset + 1,
          classes: typeof classMap[tag] === "string" ? [classMap[tag]] : classMap[tag]
        });
      }
      return lineOfCode.replace(tagRE, "");
    }).join("\n");
    return {
      code,
      lineOptions
    };
  };
}

function createTransformerDiff(options = {}) {
  const {
    classLineAdd = "diff add",
    classLineRemove = "diff remove",
    classActivePre = "has-diff"
  } = options;
  return {
    name: "plugin-shiki-transformer:diff",
    preTransformer: createRangeTransformer({
      "++": classLineAdd,
      "--": classLineRemove
    }, options),
    postTransformer: ({ code }) => {
      if (!code.includes(classLineAdd) && !code.includes(classLineRemove)) {
        return code;
      }
      return addClass(code, classActivePre, "pre");
    }
  };
}

function createTransformerFocus(options = {}) {
  const {
    classActiveLine = "focused",
    classActivePre = "has-focused"
  } = options;
  return {
    name: "plugin-shiki-transformer:focus",
    preTransformer: createRangeTransformer({
      focus: classActiveLine,
      fc: classActiveLine
    }, options),
    postTransformer: ({ code }) => {
      if (!code.includes(classActiveLine)) {
        return code;
      }
      return addClass(code, classActivePre, "pre");
    }
  };
}

function createTransformerHighlight(options = {}) {
  const {
    classActiveLine = "highlighted",
    classActivePre = "has-highlight"
  } = options;
  return {
    name: "plugin-shiki-transformer:highlight",
    preTransformer: createRangeTransformer({
      highlight: classActiveLine,
      hl: classActiveLine
    }, options),
    postTransformer: ({ code }) => {
      if (!code.includes(classActiveLine)) {
        return code;
      }
      return addClass(code, classActivePre, "pre");
    }
  };
}

function createTransformerErrorLevel(options = {}) {
  const {
    classMap = {
      error: ["highlighted", "error"],
      warning: ["highlighted", "warning"]
    },
    classActivePre = "has-highlight"
  } = options;
  return {
    name: "plugin-shiki-transformer:highlight-error",
    preTransformer: createRangeTransformer(classMap, options),
    postTransformer: ({ code }) => {
      if (!code.includes("highlighted error") && !code.includes("highlighted warning")) {
        return code;
      }
      return addClass(code, classActivePre, "pre");
    }
  };
}

const classActivePre = "has-line-number";
function createTransformerLineNumber() {
  return {
    name: "plugin-shiki-transformer:line-number",
    preTransformer: ({ code }) => {
      const lineOptions = [];
      code.split("\n").filter((v) => v !== "	").map((lineOfCode, idx) => {
        const lineNumber = idx + 1;
        lineOptions.push({
          lineOfCode,
          line: lineNumber,
          classes: ["line-number"]
        });
      });
      return {
        code,
        lineOptions
      };
    },
    postTransformer: ({ code }) => {
      return addClass(code, classActivePre, "pre");
    }
  };
}

function defineTransformer(transformer2) {
  return transformer2;
}
function transformer(transformers, code, lang) {
  return transformers.reduce((options, transformer2) => {
    const { code: code2, lineOptions } = transformer2?.preTransformer?.({
      code: options.code,
      lang
    }) ?? options;
    return {
      code: code2,
      lineOptions: [
        ...options.lineOptions,
        ...lineOptions
      ]
    };
  }, {
    code,
    lineOptions: []
  });
}
function postTransformer(transformers, code, lang) {
  return transformers.reduce((code2, transformer2) => transformer2?.postTransformer?.({
    code: code2,
    lang
  }) ?? code2, code);
}

async function getHighlighter(options = {}) {
  const highlighter = await shiki.getHighlighter(options);
  const transformers = options.transformers ?? [];
  return {
    ...highlighter,
    codeToHtml: (str, htmlOptions) => {
      const lang = typeof htmlOptions === "object" ? htmlOptions.lang : htmlOptions;
      const baseLineOptions = typeof htmlOptions === "object" ? htmlOptions.lineOptions ?? [] : [];
      const theme = typeof htmlOptions === "object" ? htmlOptions.theme : void 0;
      const { code, lineOptions } = transformer(transformers, str, lang);
      const highlighted = highlighter.codeToHtml(code, {
        lang,
        theme,
        lineOptions: [
          ...lineOptions,
          ...baseLineOptions
        ]
      });
      return postTransformer(transformers, highlighted, lang);
    }
  };
}

exports.addClass = addClass;
exports.createRangeTransformer = createRangeTransformer;
exports.createTransformerDiff = createTransformerDiff;
exports.createTransformerErrorLevel = createTransformerErrorLevel;
exports.createTransformerFocus = createTransformerFocus;
exports.createTransformerHighlight = createTransformerHighlight;
exports.createTransformerLineNumber = createTransformerLineNumber;
exports.defineTransformer = defineTransformer;
exports.getHighlighter = getHighlighter;
exports.postTransformer = postTransformer;
exports.transformer = transformer;
